#!/usr/bin/env python
import sys, os, re, stat, subprocess, tempfile

COMPUTE_PSNR = True

DURATIONS = { "st800x336": 3239.0/24, "short": 100.0/24 }

COLUMNS = ['group', 'target_bitrate', 'actual_bitrate', 'encode_time', 'decode_time']
if COMPUTE_PSNR:
    COLUMNS += ['total_psnr', 'min_psnr', 'med_psnr', 'max_psnr', 'total_ssim', 'min_ssim', 'med_ssim', 'max_ssim']

DECODER = "mplayer -really-quiet -nosound -benchmark -fps 24".split()
PSNRTOOL = "./psnrtool"

re_time = re.compile(r'(\d+\.\d+)\s*\[([12RD])\]\s*(.*)')
re_stream = re.compile(r'(.*)_(\d+)kbps\..*$')
re_psnr = re.compile(r'PSNR:\s*(\d+\.\d+)')
re_ssim = re.compile(r'SSIM:\s*(\d+\.\d+)')


class struct:
    def __init__(self, **kw):
        self.__dict__.update(kw)
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, ", ".join(["%s=%r" % i for i in self.__dict__.iteritems()]))
class SingleStream(struct):
    def get(self, key):
        value = getattr(self, key)
        if type(value) != type(0.0):
            return str(value)
        elif key.upper().endswith('SSIM'):
            return "%.5f" % value
        else:
            return "%.2f" % value


def decode_to_y4m(infile, outfile):
    if infile.find("xvid") < 0:
        return subprocess.Popen(DECODER + [ "-vo", "yuv4mpeg:file=" + outfile, infile], stdin=subprocess.PIPE)
    else:
        # we need to decode XviD with FFmpeg to avoid problems with dropped frames
        return subprocess.Popen("ffmpeg -i \"%s\" -f yuv4mpegpipe - >\"%s\" 2>/dev/null" % (infile, outfile), shell=True)


def get_psnr(stream):
    stream.total_psnr = 0.0
    stream.min_psnr = 0.0
    stream.med_psnr = 0.0
    stream.max_psnr = 0.0
    stream.total_ssim = 0.0
    stream.min_ssim = 0.0
    stream.med_ssim = 0.0
    stream.max_ssim = 0.0

    if not(os.path.isfile(stream.ref)) or not(os.path.isfile(stream.filename)):
        return

    sys.stderr.write("      %s <-> %s\r" % (stream.ref, stream.filename))
    sys.stderr.flush()
    stream.psnrfile = os.path.splitext(stream.filename)[0] + ".psnr.txt"

    tmpdir = tempfile.mkdtemp()
    ref_fifo = os.path.join(tmpdir, 'ref.y4m')
    alt_fifo = os.path.join(tmpdir, 'alt.y4m')
    os.mkfifo(ref_fifo)
    os.mkfifo(alt_fifo)
    try:
        ref_dec = decode_to_y4m(stream.ref, ref_fifo)
        alt_dec = decode_to_y4m(stream.filename, alt_fifo)
        cout, cerr = subprocess.Popen([PSNRTOOL, ref_fifo, alt_fifo, stream.psnrfile], stdout=subprocess.PIPE).communicate()
        ref_dec.wait()
        alt_dec.wait()
    finally:
        os.unlink(ref_fifo)
        os.unlink(alt_fifo)
        os.rmdir(tmpdir)

    try:
        stream.total_psnr = float(re_psnr.search(cout).group(1))
    except AttributeError:
        pass
    try:
        stream.total_ssim = float(re_ssim.search(cout).group(1))
    except AttributeError:
        pass

    try:
        f = open(stream.psnrfile, "r")
        lines = [line.strip().split("\t") for line in f if line.strip()]
        f.close()
    except (IOError, ValueError):
        return
    if not lines:
        return
    minlen = min(map(len, lines))
    if minlen > 1:
        psnrs = [float(line[1]) for line in lines]
        psnrs.sort()
        stream.min_psnr = psnrs[0]
        stream.med_psnr = psnrs[len(psnrs) / 2]
        stream.max_psnr = psnrs[-1]
    if minlen > 2:
        ssims = [float(line[2]) for line in lines]
        ssims.sort()
        stream.min_ssim = ssims[0]
        stream.med_ssim = ssims[len(ssims) / 2]
        stream.max_ssim = ssims[-1]


#if __name__ == "__main__":
#    stream = SingleStream(filename="st800x336_x264_high_2000kbps.264", ref="st800x336.mkv")
#    get_psnr(stream)
#    print
#    print stream.total_psnr, stream.total_ssim


if __name__ == "__main__":
    # parse raw time values, determine minimum of all runs
    times = {}
    for line in open("encode_times.txt"):
        m = re_time.match(line)
        if not m: continue
        t = float(m.group(1))
        mode = m.group(2)
        stream = m.group(3)
        if not stream in times:
            times[stream] = {}
        times[stream][mode] = min(times[stream].get(mode, 1.0E+100), t)

    # iterate over reference streams
    if COMPUTE_PSNR:
        print " | ".join(COLUMNS)
    results = {}
    for refstream in [s for s in times if 'R' in times[s]]:
        base = refstream.rsplit('.', 1)[0]
        ref_decode_time = times[refstream]['R']
        if base in DURATIONS:
            bitrate_factor = (1.0 / 125) / DURATIONS[base]
        else:
            bitrate_factor = 0
        for streamfile in [s for s in times if s.startswith(base) and s != refstream]:
            m = re_stream.match(streamfile)
            if not m:
                raise ValueError, "invalid stream file name: '%s'" % streamfile
            group = m.group(1)
            bitrate = int(m.group(2))
            if not group in results:
                results[group] = {}
            stream = SingleStream(filename=streamfile, target_bitrate=bitrate, group=group, ref=refstream)
            results[group][bitrate] = stream
            stream.decode_time = times[streamfile]['D']
            stream.encode_time = times[streamfile]['1'] - ref_decode_time
            if '2' in times[streamfile]:
                stream.encode_time += times[streamfile]['2'] - ref_decode_time
            try:
                stream.filesize = os.stat(streamfile)[stat.ST_SIZE]
            except OSError:
                stream.filesize = 0
            stream.actual_bitrate = bitrate_factor * stream.filesize
            if COMPUTE_PSNR:
                get_psnr(stream)
                print " | ".join([stream.get(col) for col in COLUMNS])
                sys.stdout.flush()

    # output
    out = file("encode_results.txt", "w")
    print >>out, "\t".join(COLUMNS)
    groups = results.keys()
    groups.sort()
    for group in groups:
        bitrates = results[group].keys()
        bitrates.sort()
        for bitrate in bitrates:
            stream = results[group][bitrate]
            print >>out, "\t".join([stream.get(col) for col in COLUMNS])
    out.close()
