#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define SSIM 1
#define LUMIMASK 1

////////////////////////////////////////////////////////////////////////////////

typedef struct _y4m {
    FILE *f;
    int width, height;
    uint8_t *y, *cb, *cr;
    int framesize;
} y4m_file_t;

void y4m_open(y4m_file_t *f, const char *filename) {
    char lbuf[256];
    char *lpos, *tend;
    f->f = fopen(filename, "rb");
    if (!f->f) {
        fprintf(stderr, "Error: cannot open input file '%s'\n", filename);
        exit(1);
    }
    lbuf[10] = '\0';
    lpos = fgets(lbuf, 254, f->f);
    lbuf[9] = '\0';
    if (!lpos || strcmp(lbuf, "YUV4MPEG2")) {
        fprintf(stderr, "Error: input file '%s' is not a YUV4MPEG2 file\n", filename);
        exit(1);
    }
    f->width = f->height = 0;
    tend = &lbuf[9];
    while (*(lpos = tend + 1)) {
        for (tend = lpos;  *tend && !isspace(*tend);  ++tend);
        if (!*tend) tend[1] = '\0';
        *tend = '\0';
        if (*lpos == 'W') f->width  = atoi(lpos + 1);
        if (*lpos == 'H') f->height = atoi(lpos + 1);
    }
    if ((f->width <= 0) || (f->height <= 0)) {
        fprintf(stderr, "Error: input file '%s' is invalid\n", filename);
        exit(1);
    }
    f->framesize = f->width * f->height * 3 / 2;
    f->y = malloc(f->framesize);
    if (!f->y) {
        fprintf(stderr, "Error: failed to allocate buffer for input file '%s'\n", filename);
        exit(1);
    }
    f->cb = &f->y[f->width * f->height];
    f->cr = &f->cb[f->width * f->height / 4];
}

int y4m_read_frame(y4m_file_t *f) {
    char lbuf[256];
    char *res;
    res = fgets(lbuf, 254, f->f);
    if (!res || !*lbuf)
        return 0;
    lbuf[5] = '\0';
    if (strcmp(lbuf, "FRAME")) {
        fprintf(stderr, "Error: malformed input file\n");
        exit(1);
    }
    if (fread(f->y, f->framesize, 1, f->f) != 1)
        fprintf(stderr, "Warning: unexpected end of input file\n");
    return 1;
}

void y4m_close(y4m_file_t *f) {
    fclose(f->f);
    free(f->y);
    memset(f, 0, sizeof(y4m_file_t));
}

////////////////////////////////////////////////////////////////////////////////

uint64_t plane_sqdiff(uint8_t *a, uint8_t *b, int count) {
    uint64_t sum = 0;
    while (count--) {
        uint32_t diff = (*a++) - (*b++);
        sum += diff * diff;
    }
    return sum;
}

double sqdiff2psnr(uint64_t diff, uint64_t pixels) {
    return 4.3429448190325175 * log((pixels * (255 * 255)) / (double) diff);
}

////////////////////////////////////////////////////////////////////////////////

#if SSIM

static double sum_quality = 0.0;
static double sum_plane = 0.0;
static double sum_weights = 0.0;
static double frame_weight;
static int16_t *img12_sum_block;
static int16_t* img1_sum;
static int16_t* img2_sum;
static int32_t* img1_sq_sum;
static int32_t* img2_sq_sum;
static int32_t* img12_mul_sum;
static int stride;
static int luminance;

void ssim_init(int width, int height) {
    img12_sum_block = (int16_t*) malloc(width * height / 2);
    stride = (width / 8 + 1) * 8;
    img1_sum = (int16_t*) malloc(stride * 9 * sizeof(int16_t));
    img2_sum = (int16_t*) malloc(stride * 9 * sizeof(int16_t));
    img1_sq_sum = (int32_t*) malloc(stride * 9 * sizeof(int32_t));
    img2_sq_sum = (int32_t*) malloc(stride * 9 * sizeof(int32_t));
    img12_mul_sum = (int32_t*) malloc(stride * 9 * sizeof(int32_t));
}

#define C1 (64 * 64 * 0.01*255*0.01*255)
#define C2 (64 * 64 * 0.03*255*0.03*255)
static inline double similarity(int muX, int muY, int preMuX2, int preMuY2, int preMuXY2) {
	int muX2, muY2, muXY, thetaX2, thetaY2, thetaXY;
	muX2 = muX * muX;
	muY2 = muY * muY;
	muXY = muX * muY;
	thetaX2 = 64 * preMuX2 - muX2;
	thetaY2 = 64 * preMuY2 - muY2;
	thetaXY = 64 * preMuXY2 - muXY;
	return (2 * muXY + C1) * (2 * thetaXY + C2) / ((muX2 + muY2 + C1) * (thetaX2 + thetaY2 + C2));
}

double image_quality(const uint8_t *img1, const uint8_t *img2, int width, int height) {
	int x, y, x2, y2, img1_block, img2_block, img1_sq_block, img2_sq_block, img12_mul_block, temp;
	double plane_quality, weight, mean;
	int16_t *img1_sum_ptr1, *img1_sum_ptr2;
	int16_t *img2_sum_ptr1, *img2_sum_ptr2;
	int32_t *img1_sq_sum_ptr1, *img1_sq_sum_ptr2;
	int32_t *img2_sq_sum_ptr1, *img2_sq_sum_ptr2;
	int32_t *img12_mul_sum_ptr1, *img12_mul_sum_ptr2;

    plane_quality = 0.0;
    #if LUMIMASK
        sum_plane = 0.0;
    #else
        sum_plane = (height - 7) * (width - 7);
    #endif

    temp = stride * 8;
    img1_sum_ptr1 = &img1_sum[temp];
    img2_sum_ptr1 = &img2_sum[temp];
    img1_sq_sum_ptr1 = &img1_sq_sum[temp];
    img2_sq_sum_ptr1 = &img2_sq_sum[temp];
    img12_mul_sum_ptr1 = &img12_mul_sum[temp];
	for (x = 0; x < width; x++) {
		img1_sum[x]      = img1[x];
		img2_sum[x]      = img2[x];
		img1_sq_sum[x]   = img1[x] * img1[x];
		img2_sq_sum[x]   = img2[x] * img2[x];
		img12_mul_sum[x] = img1[x] * img2[x];
		img1_sum_ptr1[x]      = 0;
		img2_sum_ptr1[x]      = 0;
		img1_sq_sum_ptr1[x]   = 0;
		img2_sq_sum_ptr1[x]   = 0;
		img12_mul_sum_ptr1[x] = 0;
	}

	for (y = 1;  y < height;  ++y) {
		img1 += width;
		img2 += width;

		temp = (y - 1) % 9 * stride;
		img1_sum_ptr1      = &img1_sum[temp];
		img2_sum_ptr1      = &img2_sum[temp];
		img1_sq_sum_ptr1   = &img1_sq_sum[temp];
		img2_sq_sum_ptr1   = &img2_sq_sum[temp];
		img12_mul_sum_ptr1 = &img12_mul_sum[temp];

		temp = y % 9 * stride;
		img1_sum_ptr2      = &img1_sum[temp];
		img2_sum_ptr2      = &img2_sum[temp];
		img1_sq_sum_ptr2   = &img1_sq_sum[temp];
		img2_sq_sum_ptr2   = &img2_sq_sum[temp];
		img12_mul_sum_ptr2 = &img12_mul_sum[temp];

		for (x = 0;  x < width;  ++x) {
			img1_sum_ptr2[x]      = img1_sum_ptr1[x] + img1[x];
			img2_sum_ptr2[x]      = img2_sum_ptr1[x] + img2[x];
			img1_sq_sum_ptr2[x]   = img1_sq_sum_ptr1[x] + img1[x] * img1[x];
			img2_sq_sum_ptr2[x]   = img2_sq_sum_ptr1[x] + img2[x] * img2[x];
			img12_mul_sum_ptr2[x] = img12_mul_sum_ptr1[x] + img1[x] * img2[x];
		}

		if (y > 6) {
			temp = (y + 1) % 9 * stride;
			img1_sum_ptr1      = &img1_sum[temp];
			img2_sum_ptr1      = &img2_sum[temp];
			img1_sq_sum_ptr1   = &img1_sq_sum[temp];
			img2_sq_sum_ptr1   = &img2_sq_sum[temp];
			img12_mul_sum_ptr1 = &img12_mul_sum[temp];

			for (x = 0;  x < width;  ++x) {
				img1_sum_ptr1[x]      = img1_sum_ptr2[x] - img1_sum_ptr1[x];
				img2_sum_ptr1[x]      = img2_sum_ptr2[x] - img2_sum_ptr1[x];
				img1_sq_sum_ptr1[x]   = img1_sq_sum_ptr2[x] - img1_sq_sum_ptr1[x];
				img2_sq_sum_ptr1[x]   = img2_sq_sum_ptr2[x] - img2_sq_sum_ptr1[x];
				img12_mul_sum_ptr1[x] = img12_mul_sum_ptr2[x] - img12_mul_sum_ptr1[x];
			}

			img1_block      = 0;
			img2_block      = 0;
			img1_sq_block   = 0;
			img2_sq_block   = 0;
			img12_mul_block = 0;

			for (x = 0; x < 8; x++) {
				img1_block      += img1_sum_ptr1[x];
				img2_block      += img2_sum_ptr1[x];
				img1_sq_block   += img1_sq_sum_ptr1[x];
				img2_sq_block   += img2_sq_sum_ptr1[x];
				img12_mul_block += img12_mul_sum_ptr1[x];
			}

            #if LUMIMASK
				y2 = y - 7;
				x2 = 0;
				if (luminance) {
					mean = (img2_block + img1_block) / 128.0;
					if (!(y2%2 || x2%2))
						*(img12_sum_block + y2/2 * width/2 + x2/2) = img2_block + img1_block;
				} else {
					mean = *(img12_sum_block + y2 * width/2 + x2);
					mean += *(img12_sum_block + y2 * width/2 + x2 + 4);
					mean += *(img12_sum_block + (y2 + 4) * width/2 + x2);
					mean += *(img12_sum_block + (y2 + 4) * width/2 + x2 + 4);
					mean /= 512.0;
				}
				weight = mean < 40 ? 0.0 : (mean < 50 ? (mean - 40.0) / 10.0 : 1.0);
				sum_plane += weight;
				plane_quality += weight * similarity(img1_block, img2_block, img1_sq_block, img2_sq_block, img12_mul_block);
            #else
                plane_quality += similarity(img1_block, img2_block, img1_sq_block, img2_sq_block, img12_mul_block);
            #endif

			for (x = 8;  x < width;  ++x) {
				img1_block      = img1_block + img1_sum_ptr1[x] - img1_sum_ptr1[x - 8];
				img2_block      = img2_block + img2_sum_ptr1[x] - img2_sum_ptr1[x - 8];
				img1_sq_block   = img1_sq_block + img1_sq_sum_ptr1[x] - img1_sq_sum_ptr1[x - 8];
				img2_sq_block   = img2_sq_block + img2_sq_sum_ptr1[x] - img2_sq_sum_ptr1[x - 8];
				img12_mul_block = img12_mul_block + img12_mul_sum_ptr1[x] - img12_mul_sum_ptr1[x - 8];

				#if LUMIMASK
					y2 = y - 7;
					x2 = x - 7;
					if (luminance) {
						mean = (img2_block + img1_block) / 128.0f;
						if (!(y2%2 || x2%2))
							*(img12_sum_block + y2/2 * width/2 + x2/2) = img2_block + img1_block;
					} else {
						mean = *(img12_sum_block + y2 * width/2 + x2);
						mean += *(img12_sum_block + y2 * width/2 + x2 + 4);
						mean += *(img12_sum_block + (y2 + 4) * width/2 + x2);
						mean += *(img12_sum_block + (y2 + 4) * width/2 + x2 + 4);
						mean /= 512.0;
					}
					weight = mean < 40 ? 0.0 : (mean < 50 ? (mean - 40.0) / 10.0 : 1.0);
					sum_plane += weight;
					plane_quality += weight * similarity(img1_block, img2_block, img1_sq_block, img2_sq_block, img12_mul_block);
				#else
					plane_quality += similarity(img1_block, img2_block, img1_sq_block, img2_sq_block, img12_mul_block);
                #endif
            }
        }
    }
    if (fabs(sum_plane) < 1.0E-6)
        return 1.0;
    else
        return plane_quality / sum_plane;
}

double ssim_frame(const uint8_t *ref, const uint8_t *alt, int width, int height) {
    double a, b, c, d;
    luminance = 1;
    a = image_quality(ref, alt, width, height);
    luminance = 0;
    frame_weight = sum_plane / ((width - 7) * (height - 7));
    if (fabs(frame_weight) < 1.0E-6) {
        a = b = c = 1.0;
    } else {
        int offset = width * height;
        b = image_quality(&ref[offset], &alt[offset], width >> 1, height >> 1);
        offset += (width * height) >> 2;
        c = image_quality(&ref[offset], &alt[offset], width >> 1, height >> 1);
    }
    d = 0.8 * a + 0.1 * (b + c);
    sum_weights += frame_weight;
    sum_quality += d * frame_weight;
    return d;
}

double ssim_done(void) {
    free(img12_sum_block);
    free(img1_sum);
    free(img2_sum);
    free(img1_sq_sum);
    free(img2_sq_sum);
    free(img12_mul_sum);
    return sum_quality / sum_weights;
}

#endif

////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[]) {
    y4m_file_t a, b;
    uint64_t frame_diff, total_diff = 0;
    int frameno, delay = 0;
    FILE *stats = NULL;

    if (argc < 3) {
        printf("Compare two YUV4MPEG streams and output per-frame and total PSNR.\n");
        printf("Usage: %s <A> <B> [<STATS> [<DELAY>]]]\n", argv[0]);
        printf("<A> is the reference stream file name,\n");
        printf("<B> is the candidate stream file name,\n");
        printf("<STATS> is the (optional) name of a text file to receive per-frame statistics.\n");
        printf("<DELAY> is the (optional) amount of delay in frames between the streams.\n");
        printf("        0 = no delay, negative = skip initial frames in A,\n");
        printf("        positive = skip initial frames in B\n");
        return 2;
    }

    y4m_open(&a, argv[1]);
    y4m_open(&b, argv[2]);
    if ((a.width != b.width) || (a.height != b.height)) {
        y4m_close(&a);
        y4m_close(&b);
        fprintf(stderr, "Error: input frame dimensions differ\n");
        return 1;
    }

    #if SSIM
        ssim_init(a.width, a.height);
    #endif

    if (argc > 3) {
        stats = fopen(argv[3], "w");
        if (!stats)
            fprintf(stderr, "Error: failed to open statistics file\n");
    }

    if (argc > 4) {
        delay = atoi(argv[4]);
        fprintf(stderr, "delay = %d\n", delay);
    }
    for (; delay < 0;  ++delay) { fprintf(stderr, "dlyframe\n"); y4m_read_frame(&a); }
    for (; delay > 0;  --delay) y4m_read_frame(&b);

    frameno = 0;
    while (y4m_read_frame(&a) && y4m_read_frame(&b)) {
        ++frameno;
        frame_diff = plane_sqdiff(a.y, b.y, a.framesize);
        total_diff += frame_diff;
        if (stats) {
            double psnr = sqdiff2psnr(frame_diff, a.framesize);
            #if SSIM
                double ssim = ssim_frame(a.y, b.y, a.width, a.height);
                fprintf(stats, "%d\t%.2f\t%.5f\t%.5f\n", frameno, psnr, ssim, frame_weight);
            #else
                fprintf(stats, "%d\t%.2f\n", frameno, psnr(frame_diff, a.framesize));
            #endif
        }
        fprintf(stderr, "\r%d", frameno);
        fflush(stderr);
    }
    fprintf(stderr, "\r");

    printf("total PSNR: %.2f dB\n", sqdiff2psnr(total_diff, a.framesize * frameno));
    #if SSIM
        printf("total SSIM: %.5f\n", ssim_done());
    #endif

    if (stats)
        fclose(stats);
    y4m_close(&a);
    y4m_close(&b);
    return 0;
}
